(*

Component name...................: SuperMask (supermsk.pas)
Classes implemented..............: none
Version..........................: 1.1.4
Status...........................: Beta
Last update......................: 2002-02-21
Author...........................: Marcello Tavares
Website..........................: http://groups.yahoo.com/group/tmail2000
Comments, bugs, suggestions to...: mycelo@yahoo.com
Language.........................: English
Platform (tested)................: Windows 95/98/98SE/2000
Requires.........................: Borland Delphi 5 Standard or better


Features
--------

1. Matching of wildcard strings, inspired on Unix regular expressions;
2. Plenty of wildcards, variable-length sets, OR sequences;
3. No classes, no assembly, no pointers nor special types, just one ordinary function!


Known Limitations
-----------------

1. Unicode strings not supported;
2. No speed optimization.


How to use
----------

  SuperMatch(source-text(string), wildcard-string(string) [, partial(boolean) [, locate(boolen) [, beggining-at(integer)]]]) -> Boolean

  when 'partial' is true, function will return true when the 'source-text' has
  characters beyond that matched by 'wildcard-string'. SM_MATCHED_CHARS global
  variable will indicate the ammount of matched characters of 'source-text'.

  when 'locate' is true, function will match starting on 'beggining-at' character
  and will shift to next position until the mask has been fitted or reached the
  end of 'source-text'; SM_MATCHED_POSITION global variable will indicate where
  the match has been succeeded; it is recommended 'partial' being true.

Supported wildcards are:

  1. Simple
  ---------

    ? - 1 any character (mandatory)
    _ - 0 or 1 any character
    * - 0 or more characters
    % - 1 or more characters
    # - 1 or more digits ('0'...'9')

  2. Sets []
  ----------

    matches a limited set of characters.

    sintax: [<occurrences>:][?|!]<starting_char>[<ending_char>][;[?|!]<starting_char>[<ending_char>]]...
    (enclosed on brackets)

    occurrences can be:

      - a number
      - a interval n1-n2
      - empty (0 or 1)
      - * (0 or more)
      - % (1 or more)

    parameters:

      ? - the character(s) must occur
      ! - the character(s) must not occur

    examples:

      [a;b]        - one mandatory occurrence of 'a' or 'b'
      [az]         - characters 'a','b','c'...'z'
      [az;09]      - characters 'a'...'z' or '0'...'9'
      [az;!mo]     - characters 'a'...'z' except 'm','n','o'
      [!ac]        - any character except 'a','b','c'
      [6:az;09]    - 6 occurrences (e.g. '9c3ax7')
      [2-5:az;09]  - 2 to 5 occurrences
      [:az]        - 0 or 1 occurrence
      [*:az;09]    - 0 or more occurrences
      [%:az;09]    - 1 or more occurrences
      [%:az;?x]    - any lowercase sequence containing an 'x'

    warning: no masks are supported inside a set.

  3. OR sequences {}
  ------------------

    matches any of the specified submasks.

    sintax: <mask_string>;<mask_string>[;<mask_string>]...
    (enclosed on braces)

    examples:

      {xyz;abcd;123}   - word 'xyz' or 'abcd' or '123'
      {[3:09];-[2:09]} - 3 digits or '-' and 2 digits
      {a*;*0}          - something starting with 'a' or ending with '0'
      {a{bc;de};fgh}   - 'abc' or 'ade' or 'fgh'

  4. Negation <>
  --------------

    match when specified submask is not true.

    sintax: <mask_string>
    (enclosed on angle brackets)

    examples:

      <{xyz;abcd;123}> - string is not 'xyz' nor 'abc' nor '123'
      <#>*             - string doesn't begin with digit(s)

    warning: unpredictable behavior when negated mask is complex.

  5. Numeric fields ()
  --------------------

    match any variable-length string respecting a numeric range.

    sintax: <minimum>;<maximum>[;<decimal-symbol>[;<other-punctuation-chars>]]
    (encolsed on parenthesis)

    parameters:

      - <minimum>, <maximum> are any Real-compatible value (including negatives),
        specifying the range where the numeric string must fit.
      - <decimal-symbol> is one character assumed to be the decimal-part
        separator (default is dot $2E).
      - <other-punctuation-chars> are characters that can be found in the number
        and must be ignored (e.g. thousand-separators, currency symbols)

    examples:

      (0;100)                - matches anything between 0 and 100 e.g. '0',
                               '25.3', '000100'
      (-15;15;,)             - matches anything between -15 and 15, assuming
                               coma as the decimal separator e.g. '-14,99'
      US$[:\s](0;100000;.;,) - matches an american dollar ammount upto 100000,
                               including comma as the thousands separator
                               e.g. 'US$ 53,982.33', 'US$150'

    warning: masks are not supported on these parameters.

  6. Escapes \
  ------------

    bypass next character (next character will not be seen as a wildcard or
    any control or reserved character).

    sintax: \<character|predefined_character|$two_digits_hex_code>

    predefined character can be:

      - s (space)
      - t (tab)
      - c (CR)
      - l (LF)

    examples:

      \*          - character '*'
      \\          - character '\'
      \s          - character space (#$20)
      \t          - character tab (#$9)
      \c          - character CR (#$0D)
      \l          - character LF (#$0A)
      \$41        - character 'A' (#$41) - hex must have 2 digits
      [1-2:\c;\l] - matches any EOL (CR, LF, CR/LF or LF/CR)
      [\[;\]]     - matches char '[' or ']'

    the following characters are reserved and must be escaped when they have to
    be interpreted as itselves

      - aways: ? _ * % # [ ] { } < > ( ) \
      - inside sets: : ; ! ? space
      - inside 'OR sequences' and 'numeric fields': ;


License stuff
-------------

SuperMask Copyleft 2001-2002

This software is provided as-is, without any express or implied
warranty. In no event will the author be held liable for any damages
arising from the use of this software.

As a freeware, the author reserve your rights to not provide support,
requested changes in the code, specific versions, improvements of any
kind and bug fixes. The main purpose is to help a little the programmers
community over the world as a whole, not just one person or organization.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented, you must not
   claim that you wrote the original software.

2. If you use this software in a product, the author musk known all details
   about its release, including objectives, availability and contact info. An
   acknowledgment in the product documentation is required.

3. Altered source versions must be plainly marked as such, and must not be
   misrepresented as being an original software. When redistributed, a copy
   must always be sent to the author.

Please, consider my hard work.


What's new in 1.1.x version
---------------------------

1.  Better documentation;
2.  Fixed unpredicable behavior of OR sequences;
3.  Nested OR sequences now working;
4.  Escapes now include CR and LF predefined characters;
5.  Range checking error fixed on set matching;
6.  Added negation submask;
7.  Added 'Partial' boolean parameter;
8.  Added SM_MATCHED_CHARS global variable;
9.  '#' mask now working;
10. Added 'Numeric Fields';
11. Added 'Locate' option and SM_MATCHED_POSITION global variable;
12. Improved set matching behavior;
13. Added mandatory occurrence on set wildcard.


Author data
-----------

Marcello Roberto Tavares Pereira
mycelo@yahoo.com
ICQ 5831833
Sorocaba/SP - BRAZIL
Spoken languages: Portuguese, English, Spanish

*)

unit supermsk;

interface

var
  SM_MATCHED_CHARS: Cardinal;
  SM_MATCHED_POSITION: Cardinal;

function SuperMatch(const Text, Mask: String; const Partial: Boolean = False; const Locate: Boolean = False; const At: Cardinal = 1): Boolean;

implementation

uses Sysutils;

const
  _SM_MAX_NESTED_SUBMASKS = 100;
  _SM_MAX_WORD_LENGTH = 20;

type
  WildcardKind = (tmFixed, tmOne, tmSeveral, tmSet, tmOrSeq, tmNegation, tmNumeric);

  PWildcardRec = ^WildcardRec;
  aWildcard = array of PWildcardRec;

  WildcardRec = record
    Sequence: String[_SM_MAX_WORD_LENGTH];
    SubMasks: array of aWildcard;
    Solid: Boolean;

    case Kind: WildcardKind of

      tmFixed, tmOne, tmSeveral, tmSet, tmOrSeq, tmNegation: (
        Valids: set of Char;
        Invalids: set of Char;
        MustHave: set of Char;
        Counter: Integer;
        Minimum: Integer;
      );

      tmNumeric: (
        MinValue: Double;
        MaxValue: Double;
        DecimalDot: Char;
      );
  end;

function SuperMatch;
var
  Root: aWildcard;
  QtdUsed: Integer;
  Buffer: PChar;
  LocatePos: Integer;

  // Normalize all escapes in hexa format

  function NormEscape(const Text: String): String;
  var
    Loop: Integer;
    Escape: Boolean;

  begin

    Loop := 1;
    Escape := False;
    Result := '';

    while Loop <= Length(Text) do
    begin

      if (not Escape) and (Text[Loop] = '\') then
      begin

        Escape := True;
        Result := Result + '\';
      end
      else
      begin

        if Escape then
        begin

          case Text[Loop] of

            '$': Result := Result + '$';
            'c': Result := Result + '$0D';
            'l': Result := Result + '$0A';
            's': Result := Result + '$20';
            't': Result := Result + '$09';

            else
              Result := Result + '$' + Format('%.2x', [Ord(Text[Loop])]);
          end;

          Escape := False;
        end
        else
        begin

          Result := Result + Text[Loop];
        end;
      end;

      Inc(Loop);
    end;
  end;

  // Replace escape with actual character

  function ConvEscape(const Text: String): String;
  var
    Loop: Integer;
    Escape: Boolean;

  begin

    Loop := 1;
    Escape := False;
    Result := '';

    while Loop <= Length(Text) do
    begin

      if (not Escape) and (Text[Loop] = '\') then
      begin

        Escape := True;
      end
      else
      begin

        if Escape and (Text[Loop] = '$')
        and (StrToIntDef(Copy(Text, Loop, 3), -1) >= 0) then
        begin

          Result := Result + Chr(StrToInt(Copy(Text, Loop, 3)));
          Inc(Loop, 2);
        end
        else
        begin

          Result := Result + Text[Loop];
        end;

        Escape := False;
      end;

      Inc(Loop);
    end;
  end;

  // Returns an item from a list

  function Item(Txt: String; Cont: Integer): String;
  var
    Quote: Integer;
    QChar: array [1.._SM_MAX_NESTED_SUBMASKS] of Char;
    Loop: Integer;
    ContItem: Integer;
    Jump: Boolean;

  begin

    Result := '';
    ContItem := 1;
    Quote := 0;
    Jump := False;

    for Loop := 1 to Length(Txt) do
    begin

      if Txt[Loop] in ['{', '[', '<', '('] then
      begin

        Inc(Quote);

        case Txt[Loop] of

          '{': QChar[Quote] := '}';
          '[': QChar[Quote] := ']';
          '<': QChar[Quote] := '>';
          '(': QChar[Quote] := ')';
        end;
      end
      else
      begin

        if (Quote > 0) and (Txt[Loop] = QChar[Quote]) then
        begin

          Dec(Quote);
        end
        else
        begin

          if (Quote = 0) and (Txt[Loop] = ';') then
          begin

            Inc(ContItem);
            Jump := True;
          end;
        end;
      end;

      if (not Jump) and (ContItem = Cont) then
      begin

        Result := Result + Txt[Loop];
      end;

      Jump := False;
    end;
  end;

  // Create a new wildcard structure

  procedure NewWildcard(var Wildcards: aWildcard);
  begin

    SetLength(Wildcards, High(Wildcards)+2);
    New(Wildcards[High(Wildcards)]);
    Wildcards[High(Wildcards)]^.Valids := [];
    Wildcards[High(Wildcards)]^.Invalids := [];
    Wildcards[High(Wildcards)]^.MustHave := [];
    Wildcards[High(Wildcards)]^.Counter := 0;
    Wildcards[High(Wildcards)]^.Solid := False;
    Wildcards[High(Wildcards)]^.Sequence := '';
    Wildcards[High(Wildcards)]^.SubMasks := nil;
  end;

  // Create a new set []

  procedure NewSet(var Wildcards: aWildcard; Txt: String);
  var
    Loop, nPos: Integer;
    Charac: Byte;
    StrCont: String;
    Range: String;
    RangeType: Char;

  begin

    if Txt = '' then
      Exit;

    NewWildcard(Wildcards);
    Wildcards[High(Wildcards)]^.Kind := tmSet;
    Wildcards[High(Wildcards)]^.Solid := True;
    Wildcards[High(Wildcards)]^.Counter := 1;
    Wildcards[High(Wildcards)]^.Minimum := 1;

    if Pos(':', Txt) > 0 then
    begin

      Loop := 1;
      StrCont := '';

      while (Txt[Loop] <> ':') and (Loop <= Length(Txt)) do
      begin

        StrCont := StrCont + Txt[Loop];
        Inc(Loop);
      end;

      if StrCont = '' then
      begin

        Wildcards[High(Wildcards)]^.Counter := 1;
        Wildcards[High(Wildcards)]^.Minimum := 0;
        Wildcards[High(Wildcards)]^.Solid := False;
      end
      else
      begin

        case StrCont[1] of

          '%':
          begin

            Wildcards[High(Wildcards)]^.Counter := -1;
            Wildcards[High(Wildcards)]^.Minimum := -1;
            Wildcards[High(Wildcards)]^.Solid := True;
          end;

          '*':
          begin

            Wildcards[High(Wildcards)]^.Counter := -1;
            Wildcards[High(Wildcards)]^.Minimum := -1;
            Wildcards[High(Wildcards)]^.Solid := False;
          end;

          else
          begin

            nPos := Pos('-', StrCont);

            if nPos > 0 then
            begin

              Wildcards[High(Wildcards)]^.Minimum := StrToIntDef(Copy(StrCont, 1, nPos-1), 1);
              Wildcards[High(Wildcards)]^.Counter := StrToIntDef(Copy(StrCont, nPos+1, Length(StrCont)), 1);
            end
            else
            begin

              Wildcards[High(Wildcards)]^.Minimum := StrToIntDef(StrCont, 1);
              Wildcards[High(Wildcards)]^.Counter := StrToIntDef(StrCont, 1);
            end;

            Wildcards[High(Wildcards)]^.Solid := Wildcards[High(Wildcards)]^.Minimum > 0;
          end;
        end;
      end;

      Txt := Copy(Txt, Loop+1, Length(Txt));
    end;

    Loop := 0;

    repeat

      Inc(Loop);
      Range := Item(Txt, Loop);

      if Length(Range) > 0 then
      begin

        if Range[1] in ['!', '?'] then
          RangeType := Range[1]
        else
          RangeType := #0;

        if RangeType <> #0 then
          Range := ConvEscape(Copy(Range, 2, Length(Range)-1))
        else
          Range := ConvEscape(Range);

        if Length(Range) = 1 then
        begin

          case RangeType of

            '!': Wildcards[High(Wildcards)]^.Invalids := Wildcards[High(Wildcards)]^.Invalids + [Range[1]];

            '?':
            begin

              Wildcards[High(Wildcards)]^.Valids := Wildcards[High(Wildcards)]^.Valids + [Range[1]];
              Wildcards[High(Wildcards)]^.MustHave := Wildcards[High(Wildcards)]^.MustHave + [Range[1]];
            end;

            else
              Wildcards[High(Wildcards)]^.Valids := Wildcards[High(Wildcards)]^.Valids + [Range[1]];
          end;
        end;

        if (Length(Range) = 2) and (Ord(Range[1]) < Ord(Range[2])) then
        begin

          for Charac := Ord(Range[1]) to Ord(Range[2]) do
          begin

            case RangeType of

              '!': Wildcards[High(Wildcards)]^.Invalids := Wildcards[High(Wildcards)]^.Invalids + [Chr(Charac)];

              '?':
              begin

                Wildcards[High(Wildcards)]^.Valids := Wildcards[High(Wildcards)]^.Valids + [Chr(Charac)];
                Wildcards[High(Wildcards)]^.MustHave := Wildcards[High(Wildcards)]^.MustHave + [Chr(Charac)];
              end;

              else
                Wildcards[High(Wildcards)]^.Valids := Wildcards[High(Wildcards)]^.Valids + [Chr(Charac)];
            end;
          end;
        end;
      end;

    until Length(Range) = 0;
  end;

  // Create a new numeric field

  procedure NewNumField(var Wildcards: aWildcard; Txt: String);
  var
    Loop: Byte;
    nPos: Integer;
    Str: String;
    Ok: Boolean;
    Min, Max: Double;
    Dot: Char;
    Punct: String;

  begin

    Ok := True;
    Min := 0;
    Max := -1;
    Dot := #0;

    for Loop := 1 to 4 do
    begin

      Str := ConvEscape(Item(Txt, Loop));

      case Loop of

        1:
        begin

          Val(Str, Min, nPos);
          Ok := nPos = 0;
        end;

        2:
        begin

          Val(Str, Max, nPos);
          Ok := nPos = 0;
        end;

        3:
        begin

          Ok := Length(Str) <= 1;

          if Ok then
          begin

            if Length(Str) = 0 then
              Dot := '.'
            else
              Dot := Str[1];
          end;
        end;

        4:
        begin

          Punct := Str;
        end;
      end;

      if not Ok then
        Break;
    end;

    if Ok then
    begin

      NewWildcard(Wildcards);
      Wildcards[High(Wildcards)]^.Kind := tmNumeric;
      Wildcards[High(Wildcards)]^.Solid := True;
      Wildcards[High(Wildcards)]^.MinValue := Min;
      Wildcards[High(Wildcards)]^.MaxValue := Max;
      Wildcards[High(Wildcards)]^.DecimalDot := Dot;
      Wildcards[High(Wildcards)]^.Sequence := Punct;
    end;
  end;

  // Divides the mask in wildcards and sets

  procedure Divide(var Wildcards: aWildcard; Txt: String);
  var
    Loop, Loop2: Integer;
    Wildset, NumField: Boolean;
    Group: Integer;
    Str: String;
    SubTxt: String;
    EndChar: array[1.._SM_MAX_NESTED_SUBMASKS] of Char;

  begin

    SetLength(Wildcards, 0);
    Wildset := False;
    NumField := False;
    Group := 0;
    Str := '';

    for Loop := 1 to Length(Txt)+1 do
    begin

      if Group = 0 then
      begin

        if not Wildset then
        begin

          if not NumField then
          begin

            if ((Loop > Length(Txt))
            or (Txt[Loop] in ['*', '?', '_', '%', '#', '[', '{', '<', '('])) then
            begin

              if Str <> '' then
              begin

                NewWildcard(Wildcards);

                Wildcards[High(Wildcards)]^.Kind := tmFixed;
                Wildcards[High(Wildcards)]^.Sequence := ConvEscape(Str);
                Wildcards[High(Wildcards)]^.Solid := True;
                Wildcards[High(Wildcards)]^.Counter := Length(Wildcards[High(Wildcards)]^.Sequence);

                Str := '';
              end;

              if Loop <= Length(Txt) then
              begin

                case Txt[Loop] of

                  '*':
                  begin

                    NewWildcard(Wildcards);

                    Wildcards[High(Wildcards)]^.Kind := tmSeveral;
                    Wildcards[High(Wildcards)]^.Solid := False;
                  end;

                  '%':
                  begin

                    NewWildcard(Wildcards);

                    Wildcards[High(Wildcards)]^.Kind := tmSeveral;
                    Wildcards[High(Wildcards)]^.Solid := True;
                  end;

                  '#':
                  begin

                    NewWildcard(Wildcards);

                    Wildcards[High(Wildcards)]^.Kind := tmSet;
                    Wildcards[High(Wildcards)]^.Solid := True;
                    Wildcards[High(Wildcards)]^.Valids := ['0'..'9'];
                    Wildcards[High(Wildcards)]^.Invalids := [];
                    Wildcards[High(Wildcards)]^.Counter := -1;
                    Wildcards[High(Wildcards)]^.Minimum := -1;
                  end;

                  '?':
                  begin

                    NewWildcard(Wildcards);

                    Wildcards[High(Wildcards)]^.Kind := tmOne;
                    Wildcards[High(Wildcards)]^.Solid := True;
                  end;

                  '_':
                  begin

                    NewWildcard(Wildcards);

                    Wildcards[High(Wildcards)]^.Kind := tmOne;
                    Wildcards[High(Wildcards)]^.Solid := False;
                  end;

                  '[':
                  begin

                    Wildset := True;
                  end;

                  '(':
                  begin

                    NumField := True;
                  end;

                  '{':
                  begin

                    Inc(Group);
                    EndChar[Group] := '}';
                  end;

                  '<':
                  begin

                    Inc(Group);
                    EndChar[Group] := '>';
                  end;
                end;
              end;
            end  // wildcard
            else
            begin

              Str := Str + Txt[Loop];
            end;
          end  // numeric field
          else
          begin

            if Txt[Loop] = ')' then
            begin

              NumField := False;
              NewNumField(Wildcards, Str);
              Str := '';
            end
            else
            begin

              Str := Str + Txt[Loop];
            end;
          end;
        end  // set
        else
        begin

          if Txt[Loop] = ']' then
          begin

            Wildset := False;
            NewSet(Wildcards, Str);
            Str := '';
          end
          else
          begin

            Str := Str + Txt[Loop];
          end;
        end;
      end  // Group
      else
      begin

        case Txt[Loop] of

          '{':
          begin

            Inc(Group);
            EndChar[Group] := '}';
          end;

          '<':
          begin

            Inc(Group);
            EndChar[Group] := '>';
          end;

          else
            if Txt[Loop] = EndChar[Group] then
              Dec(Group);
        end;

        if Group = 0 then
        begin

          case Txt[Loop] of

            '}': // OR Sequence
            begin

              NewWildcard(Wildcards);
              Wildcards[High(Wildcards)]^.Kind := tmOrSeq;
              Wildcards[High(Wildcards)]^.Solid := True;

              Loop2 := 1;
              SubTxt := Item(Str, Loop2);

              while SubTxt <> '' do
              begin

                SetLength(Wildcards[High(Wildcards)]^.SubMasks, Loop2);
                Divide(Wildcards[High(Wildcards)]^.SubMasks[Loop2-1], SubTxt);
                Inc(Loop2);
                SubTxt := Item(Str, Loop2);
              end;
            end;

            '>': // Negation
            begin

              NewWildcard(Wildcards);
              Wildcards[High(Wildcards)]^.Kind := tmNegation;
              Wildcards[High(Wildcards)]^.Solid := True;
              SetLength(Wildcards[High(Wildcards)]^.SubMasks, 1);
              Divide(Wildcards[High(Wildcards)]^.SubMasks[0], Str);
            end;
          end;

          Str := '';
        end
        else
        begin

          Str := Str + Txt[Loop];
        end;
      end;
    end; // loop
  end;

  // Recursive wildcard matching

  function Matching(var Wildcards: aWildcard; WCIndex: Integer; Txt: String; var Qtd: Integer; SubMask: Boolean; Negation: Boolean; Partial: Boolean): Boolean;
  var
    Loop: Integer;
    QtdUsed: Integer;
    StrValue: String;
    NumValue: Double;
    Code: Integer;
    Ok: Boolean;
    MustHaveList: set of Char;

  begin

    if WCIndex > High(Wildcards) then
    begin

      Result := SubMask or (Txt = '') or Partial;
      Exit;
    end;

    if Txt = '' then
    begin

      Result := (not Wildcards[WCIndex]^.Solid) and Matching(Wildcards, WCIndex+1, '', Qtd, SubMask, Negation, Partial);
      Exit;
    end;

    Result := False;

    with Wildcards[WCIndex]^ do
    begin

      case Kind of

        tmFixed:
        begin

          if Copy(Txt, 1, Counter) = Sequence then
            Result := Matching(Wildcards, WCIndex+1, Copy(Txt, Counter+1, Length(Txt)), Qtd, SubMask, Negation, Partial);

          if Result xor Negation then
            Inc(Qtd, Counter);
        end;

        tmOne:
        begin

          Result := Matching(Wildcards, WCIndex+1, Copy(Txt, 2, Length(Txt)), Qtd, SubMask, Negation, Partial);

          if Result xor Negation then
            Inc(Qtd);

          if (not Result) and (not Solid) then
          begin

            Result := Matching(Wildcards, WCIndex+1, Txt, Qtd, SubMask, Negation, Partial);
          end;
        end;

        tmSeveral:
        begin

          if Solid then
            Loop := 2
          else
            Loop := 1;

          Result := False;

          while (not Result) and (Loop <= Length(Txt)+1) do
          begin

            Result := Result or Matching(Wildcards, WCIndex+1, Copy(Txt, Loop, Length(Txt)), Qtd, SubMask, Negation, Partial{ and (Loop = Length(Txt)+1)});
            Inc(Loop);
          end;

          if Result xor Negation then
            Inc(Qtd, Loop-2);
        end;

        tmSet:
        begin

          if Solid and (Minimum > Length(Txt)) then
          begin

            Result := False;
          end
          else
          begin

            if Counter > 0 then
            begin

              Loop := 1;
              Result := False;
              Ok := True;
              MustHaveList := MustHave;

              while (not Result) and (Loop <= Counter) and (Loop <= Length(Txt)) do
              begin

                if Valids <> [] then
                  Ok := Ok and (Txt[Loop] in Valids) and (not (Txt[Loop] in Invalids))
                else
                  Ok := Ok and (not (Txt[Loop] in Invalids));

                MustHaveList := MustHaveList - [Txt[Loop]];

                Result := Ok and (Loop >= Minimum) and (MustHaveList = []) and Matching(Wildcards, WCIndex+1, Copy(Txt, Loop+1, Length(Txt)), Qtd, SubMask, Negation, Partial);

                if not Result then
                  Inc(Loop);
              end;

{             if not Result then
                Result := (Loop - 1) >= Minimum;
              Result := Result and Matching(Wildcards, WCIndex+1, Copy(Txt, Loop, Length(Txt)), Qtd, SubMask, Negation, Partial);}
            end
            else
            begin

              Loop := 1;
              Result := False;
              Ok := True;
              MustHaveList := MustHave;

              while (not Result) and (Loop <= Length(Txt)) do
              begin

                if Valids <> [] then
                  Ok := Ok and (Txt[Loop] in Valids) and (not (Txt[Loop] in Invalids))
                else
                  Ok := Ok and (not (Txt[Loop] in Invalids));

                MustHaveList := MustHaveList - [Txt[Loop]];

                Result := Ok and (MustHaveList = []) and Matching(Wildcards, WCIndex+1, Copy(Txt, Loop+1, Length(Txt)), Qtd, SubMask, Negation, Partial);

                if not Result then
                  Inc(Loop);
              end;

{             if Solid and (Loop = 1) then
                Result := False
              else
                Result := Matching(Wildcards, WCIndex+1, Copy(Txt, Loop, Length(Txt)), Qtd, SubMask, Negation, Partial);}
            end;

            if Result xor Negation then
              Inc(Qtd, Loop);

            if (not Solid) and (not Result) then
              Result := Matching(Wildcards, WCIndex+1, Copy(Txt, 1, Length(Txt)), Qtd, SubMask, Negation, Partial);
          end;
        end;

        tmOrSeq:
        begin

          Result := False;

          for Loop := Low(SubMasks) to High(SubMasks) do
          begin

            QtdUsed := 0;

            if Matching(SubMasks[Loop], 0, Txt, QtdUsed, SubMask or (WCIndex < High(Wildcards)), Negation, Partial) then
              Result := Matching(Wildcards, WCIndex+1, Copy(Txt, QtdUsed+1, Length(Txt)), Qtd, SubMask, Negation, Partial);

            if Result then
              Break;
          end;

          if Result xor Negation then
            Inc(Qtd, QtdUsed);
        end;

        tmNegation:
        begin

          Result := False;
          QtdUsed := 0;

          if not Matching(SubMasks[0], 0, Txt, QtdUsed, SubMask or (WCIndex < High(Wildcards)), not Negation, Partial) then
            Result := Matching(Wildcards, WCIndex+1, Copy(Txt, QtdUsed+1, Length(Txt)), Qtd, SubMask, Negation, Partial);

          if Result xor Negation then
            Inc(Qtd, QtdUsed);
        end;

        tmNumeric:
        begin

          Loop := 0;
          Ok := True;
          StrValue := '';

          while Ok and (Loop < Length(Txt)) do
          begin

            Ok := Ok and ((Txt[Loop+1] in ['0'..'9', '-', DecimalDot]) or (Pos(Txt[Loop+1], Sequence) > 0));

            if Ok then
            begin

              if Txt[Loop+1] in ['0'..'9', '-'] then
                StrValue := StrValue + Txt[Loop+1]
              else
                if Txt[Loop+1] = DecimalDot then
                  StrValue := StrValue + '.';
            end;

            if Ok then
              Inc(Loop);
          end;

          Val(StrValue, NumValue, Code);

          Result := (Loop > 0) and (Code = 0) and (StrValue <> '') and (NumValue >= MinValue) and (NumValue <= MaxValue);
          Result := Result and Matching(Wildcards, WCIndex+1, Copy(Txt, Loop+1, Length(Txt)), Qtd, SubMask, Negation, Partial);

          if Result xor Negation then
            Inc(Qtd, Loop);
        end;
      end;
    end;
  end;

  // Release structure memory

  procedure RelMemory(var Wildcards: aWildcard);
  var
    Loop1, Loop2: Integer;

  begin

    for Loop1 := Low(Wildcards) to High(Wildcards) do
    begin

      for Loop2 := Low(Wildcards[Loop1]^.SubMasks) to High(Wildcards[Loop1]^.SubMasks) do
        RelMemory(Wildcards[Loop1]^.SubMasks[Loop2]);

      SetLength(Wildcards[Loop1]^.SubMasks, 0);
      Dispose(Wildcards[Loop1]);
    end;

    SetLength(Wildcards, 0);
  end;

begin

  SetLength(Root, 0);

  SM_MATCHED_CHARS := 0;
  SM_MATCHED_POSITION := 0;
  Result := False;

  // Search for wildcards
  Divide(Root, NormEscape(Mask));

  // Match text with wildcards

  if Locate then
  begin

    GetMem(Buffer, Length(Text));
    StrPCopy(Buffer, Text);
    LocatePos := At-1;

    while Buffer[LocatePos] <> #0 do
    begin

      QtdUsed := 0;

      if Matching(Root, 0, String(Buffer+LocatePos), QtdUsed, False, False, Partial) then
      begin

        SM_MATCHED_POSITION := LocatePos + 1;
        SM_MATCHED_CHARS := QtdUsed;
        Result := True;
        Break;
      end;

      Inc(LocatePos);
    end;

    FreeMem(Buffer);
  end
  else
  begin

    QtdUsed := 0;
    Result := Matching(Root, 0, Copy(Text, At, Length(Text)), QtdUsed, False, False, Partial);
    SM_MATCHED_CHARS := QtdUsed;
  end;

  // Release memory
  RelMemory(Root);
end;

end.
